CON

  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000

  LF            = 10
  CR            = 13
  QU            = 34

  GPS_RCV       = 8             'Serial input from GPS.
  GPS_XMT       = 9             'Serial output to GPS.
  XB_RCV        = 2             'Serial input from XBEE.
  XB_XMT        = 3             'Serial output to XBEE.
  STS_LED       = 26            'Status LED.

  HDOP_THLD     = 110

OBJ

  csv   : "CsvFile"
  gps   : "FullDuplexSerial"

VAR

  long longitude, latitude, zero
  byte gps_msg[6]
  byte gps_utc[16], gps_stat[16], gps_lat[16], gps_ns[16], gps_lon[16]
  byte gps_ew[16], gps_spd[16], gps_course[16], gps_date[16]
  
  byte gps_auto[16], gps_3d[16], gps_prns[16 * 12], gps_dop[16], gps_hdop[16], gps_vdop[16]

PUB start | hdop

  gps.start(GPS_RCV, XB_XMT, %0000, 9600)               'Serial I/O receives from GPS, transmits to XBEE.
  dira[STS_LED]~~
  dira[GPS_XMT]~~                                       'Set output pin to GPS to "mark."
  outa[GPS_XMT]~~                                       '(Delete these two statements if GPS_XMT used with FDS.)
  csv.start
  csv.openWrite(string("track.kml"))                    'Open the KML file,
  csv.writeStr(@prologue)                               ' and write the prologue.
  
  repeat                                                'Wait for GPS status to change to "A"
    get_gps(string("GPRMC"), @gps_utc, 9)
  until (gps_stat == "A")
  
  repeat                                                'Wait for HDOP to be low enough to start.
    get_gps(string("GPGSA"), @gps_auto, 17)
    gps.str(@gps_hdop)
    gps.tx(13)
    hdop := get_integer(@gps_hdop, 0, 0)
  until (hdop < HDOP_THLD)
  
  outa[STS_LED]~~                                       'Turn on status LED.
                                        
  repeat 120                                            'Get 120 readings, then close file and quit.
    get_gps(string("GPRMC"), @gps_utc, 9)
    latitude := get_udegrees(@gps_lat, gps_ns)
    longitude := get_udegrees(@gps_lon, gps_ew)
    if (latitude)
      csv.writeCsvLine(@longitude, string("660"))
      
  csv.writeStr(@epilogue)                               'Write the epilogue to file,
  csv.close                                             '  and close it.
  outa[STS_LED]~                                        'Turn off status LED.
  repeat

PRI get_gps(msg, field_addr, field_cnt) | i, ch, field

  '' Get the GPS message matching msg.

  repeat
    repeat
    until gps.rx == "$"                                 'Wait for a $.
    repeat i from 0 to 4                                'Read the message ID.
      gps_msg[i] := gps.rx
    if strcomp(@gps_msg, msg)                           'Is it the desired message?
      gps.rx                                            '  Yes: Get past the first comma.
      repeat field from 0 to field_cnt - 1              '       Get field_cnt fields of data.
        i := 0                                          '       Point to the first character in the field.
        repeat until (ch := gps.rx) == ","              '       Keep buffering up to the next comma.
          byte[field_addr][field * 16 + i++] := ch      '       Buffer the character in the field and bump i.
        byte[field_addr][field * 16 + i] := 0           '       Buffer a zero in the field for end of string.
      return                                          '       Done.
   
PRI get_udegrees(str_addr, dir) | length, minute_pos, minutes, degrees

  '' Convert GPS ddmm.mmmmm or dddmm.mmmmm to microdegrees.

  length := strsize(str_addr)
  minute_pos := length - 8
  minutes := get_integer(str_addr, minute_pos, 0)
  degrees := get_integer(str_addr, 0, length - 8)
  degrees *= 1_000_000
  degrees += minutes / 6
  if (dir == "S" or dir == "W")
    -degrees
  return degrees

PRI get_integer(str_addr, pos, length) | i, ch

  '' Convert string containing digits and decimal point to integer (ignoring decimal point).

  i := pos                                'Start at pos.
  repeat                                  'Repeat ...
    ch := byte[str_addr][i]             '  Get the next character from the string.
    if (ch => "0" and ch =< "9")        '  Is it a digit?
      result := result * 10 + ch - "0"    '    Yes: Accumulate it into the result.
    elseifnot (ch == ".")               '  Else is it a decimal point?
      return                              '    No:  We're done. Return the result.
    if (length and ++i - pos => length)   '  Is length non-zero and ++i - pos => length?
      return                              '    Yes: We're done. Return the result.

DAT

'' Minimal KML prologue and epilogue strings (without indentation).

prologue      byte      "<?xml version=", QU, "1.0", QU, " encoding=", QU, "UTF-8", QU, "?>", CR, LF
              byte      "<kml xmlns=", QU, "http://www.opengis.net/kml/2.2", QU, ">", CR, LF
              byte      "<Document>", CR, LF
              byte      "<name>Paths</name>", CR, LF
              byte      "<Style id=", QU, "RedLine", QU, ">", CR, LF
              byte      "<LineStyle>", CR, LF
              byte      "<color>ff0000ff</color>", CR, LF
              byte      "<width>4</width>", CR, LF
              byte      "</LineStyle>", CR, LF
              byte      "</Style>", CR, LF
              byte      "<Placemark>", CR, LF
              byte      "<styleUrl>#RedLine</styleUrl>", CR, LF
              byte      "<LineString>", CR, LF
              byte      "<altitudeMode>relative</altitudeMode>", CR, LF
              byte      "<coordinates>", CR, LF, 0

epilogue      byte      "</coordinates>", CR, LF
              byte      "</LineString>", CR, LF
              byte      "</Placemark>", CR, LF
              byte      "</Document>", CR, LF
              byte      "</kml>", CR, LF, 0
